# GromacsWrapper: utilities.py
# Copyright (c) 2009 Oliver Beckstein <orbeckst@gmail.com>
# Released under the GNU Public License 3 (or higher, your choice)
# See the file COPYING for details.

"""
:mod:`gromacs.utilities` -- Helper functions and classes
========================================================

The module defines some convenience functions and classes that are
used in other modules; they do *not* make use of :mod:`gromacs.tools`
or :mod:`gromacs.cbook` and can be safely imported at any time.


Classes
-------

:class:`FileUtils` provides functions related to filename handling. It
can be used as a base or mixin class. The :class:`gromacs.analysis.Simulation`
class is derived from it.

.. autoclass:: FileUtils
   :members:
.. autoclass:: AttributeDict
.. autoclass:: Timedelta

Functions
---------

Some additional convenience functions that deal with files and
directories:

.. function:: openany(directory[,mode='r'])

   Context manager to open a compressed (bzip2, gzip) or plain file
   (uses :func:`anyopen`).

.. autofunction:: anyopen
.. autofunction:: isstream
.. autofunction:: realpath
.. function:: in_dir(directory[,create=True])

   Context manager to execute a code block in a directory.

   * The *directory* is created if it does not exist (unless
     *create* = ``False`` is set)
   * At the end or after an exception code always returns to
     the directory that was the current directory before entering
     the block.

.. autofunction:: find_first
.. autofunction:: withextsep
.. autofunction:: which

Functions that improve list processing and which do *not* treat
strings as lists:

.. autofunction:: iterable
.. autofunction:: asiterable
.. autofunction:: firstof

Functions that help handling Gromacs files:

.. autofunction:: unlink_f
.. autofunction:: unlink_gmx
.. autofunction:: unlink_gmx_backups
.. autofunction:: number_pdbs

Functions that make working with matplotlib_ easier:

.. _matplotlib: http://matplotlib.sourceforge.net/

.. autofunction:: activate_subplot
.. autofunction:: remove_legend


Miscellaneous functions:

.. autofunction:: convert_aa_code
.. autofunction:: autoconvert
.. autofunction:: hasmethod

Data
----

.. autodata:: amino_acid_codes

"""
from __future__ import absolute_import, with_statement, division

__docformat__ = "restructuredtext en"

import six
from six import string_types

import sys
import os
import glob
import fnmatch
import itertools
import re
import warnings
import errno
import subprocess
from contextlib import contextmanager
import bz2
import gzip
import datetime

import numpy

import logging

logger = logging.getLogger("gromacs.utilities")

from .exceptions import AutoCorrectionWarning


class AttributeDict(dict):
    """A dictionary with pythonic access to keys as attributes --- useful for interactive work."""

    def __getattribute__(self, x):
        try:
            return super(AttributeDict, self).__getattribute__(x)
        except AttributeError:
            return self[x]

    def __setattr__(self, name, value):
        try:
            super(AttributeDict, self).__setitem__(name, value)
        except KeyError:
            super(AttributeDict, self).__setattr__(name, value)

    def __getstate__(self):
        return self

    def __setstate__(self, state):
        self.update(state)


def autoconvert(s):
    """Convert input to a numerical type if possible.

    1. A non-string object is returned as it is
    2. Try conversion to int, float, str.
    """
    if type(s) is not str:
        return s
    for converter in int, float, str:  # try them in increasing order of lenience
        try:
            s = [converter(i) for i in s.split()]
            if len(s) == 1:
                return s[0]
            else:
                return numpy.array(s)
        except (ValueError, AttributeError):
            pass
    raise ValueError("Failed to autoconvert {0!r}".format(s))


@contextmanager
def openany(datasource, mode="rt", reset=True):
    """Context manager for :func:`anyopen`.

    Open the `datasource` and close it when the context of the :keyword:`with`
    statement exits.

    `datasource` can be a filename or a stream (see :func:`isstream`). A stream
    is reset to its start if possible (via :meth:`~io.IOBase.seek` or
    :meth:`~cString.StringIO.reset`).

    The advantage of this function is that very different input sources
    ("streams") can be used for a "file", ranging from files on disk (including
    compressed files) to open file objects to sockets and strings---as long as
    they have a file-like interface.

    :Arguments:
      *datasource*
           a file or a stream
      *mode*
           {'r', 'w'} (optional), open in r(ead) or w(rite) mode
      *reset*
           bool (optional) try to read (`mode` 'r') the stream from the
           start [``True``]


    **Example**

    Open a gzipped file and process it line by line::

        with openany("input.pdb.gz") as pdb:
            for line in pdb:
                if line.startswith('ATOM'):
                    print(line)

    Open a URL and read it::

       import urllib2
       with openany(urllib2.urlopen("https://www.mdanalysis.org/")) as html:
           print(html.read())


    .. SeeAlso::
       :func:`anyopen`

    """
    stream = anyopen(datasource, mode=mode, reset=reset)
    try:
        yield stream
    finally:
        stream.close()


# On python 3, we want to use bz2.open to open and uncompress bz2 files. That
# function allows to specify the type of the uncompressed file (bytes ot text).
# The function does not exist in python 2, thus we must use bz2.BZFile to
# which we cannot tell if the uncompressed file contains bytes or text.
# Therefore, on python 2 we use a proxy function that removes the type of the
# uncompressed file from the `mode` argument.
try:
    bz2.open
except AttributeError:
    # We are on python 2 and bz2.open is not available
    def bz2_open(filename, mode):
        """Open and uncompress a BZ2 file"""
        mode = mode.replace("t", "").replace("b", "")
        return bz2.BZ2File(filename, mode)

else:
    # We are on python 3 so we can use bz2.open
    bz2_open = bz2.open


def anyopen(datasource, mode="rt", reset=True):
    """Open datasource (gzipped, bzipped, uncompressed) and return a stream.

    `datasource` can be a filename or a stream (see :func:`isstream`). By
    default, a stream is reset to its start if possible (via
    :meth:`~io.IOBase.seek` or :meth:`~cString.StringIO.reset`).

    If possible, the attribute ``stream.name`` is set to the filename or
    "<stream>" if no filename could be associated with the *datasource*.

    :Arguments:
      *datasource*
        a file (from :class:`file` or :func:`open`) or a stream (e.g. from
        :func:`urllib2.urlopen` or :class:`cStringIO.StringIO`)
      *mode*
        {'r', 'w', 'a'} (optional),
        Open in r(ead), w(rite) or a(ppen) mode. More complicated
        modes ('r+', 'w+', ...) are not supported; only the first letter of
        `mode` is used and thus any additional modifiers are silently ignored.
      *reset*
        bool (optional),
        try to read (`mode` 'r') the stream from the start

    :Returns:
       file-like object


    .. SeeAlso::
       :func:`openany` to be used with the :keyword:`with` statement.

    """
    handlers = {"bz2": bz2_open, "gz": gzip.open, "": open}

    if mode.startswith("r"):
        if isstream(datasource):
            stream = datasource
            try:
                filename = str(stream.name)  # maybe that does not always work?
            except AttributeError:
                filename = "<stream>"
            if reset:
                try:
                    stream.reset()
                except (AttributeError, IOError):
                    try:
                        stream.seek(0)
                    except (AttributeError, IOError):
                        warnings.warn(
                            "Stream {0}: not guaranteed to be at the beginning."
                            "".format(filename),
                            category=StreamWarning,
                        )
        else:
            stream = None
            filename = datasource
            for ext in ("bz2", "gz", ""):  # file == '' should be last
                openfunc = handlers[ext]
                stream = _get_stream(datasource, openfunc, mode=mode)
                if stream is not None:
                    break
            if stream is None:
                raise IOError(
                    errno.EIO,
                    "Cannot open file or stream in mode={mode!r}.".format(**vars()),
                    repr(filename),
                )
    elif mode.startswith("w") or mode.startswith("a"):  # append 'a' not tested...
        if isstream(datasource):
            stream = datasource
            try:
                filename = str(stream.name)  # maybe that does not always work?
            except AttributeError:
                filename = "<stream>"
        else:
            stream = None
            filename = datasource
            name, ext = os.path.splitext(filename)
            if ext.startswith("."):
                ext = ext[1:]
            if not ext in ("bz2", "gz"):
                ext = ""  # anything else but bz2 or gz is just a normal file
            openfunc = handlers[ext]
            stream = openfunc(datasource, mode=mode)
            if stream is None:
                raise IOError(
                    errno.EIO,
                    "Cannot open file or stream in mode={mode!r}.".format(**vars()),
                    repr(filename),
                )
    else:
        raise NotImplementedError(
            "Sorry, mode={mode!r} is not implemented for {datasource!r}".format(
                **vars()
            )
        )
    try:
        stream.name = filename
    except (AttributeError, TypeError):
        pass  # can't set name (e.g. cStringIO.StringIO)
    return stream


def _get_stream(filename, openfunction=open, mode="r"):
    """Return open stream if *filename* can be opened with *openfunction* or else ``None``."""
    try:
        stream = openfunction(filename, mode=mode)
    except (IOError, OSError) as err:
        # An exception might be raised due to two reasons, first the openfunction is unable to open the file, in this
        # case we have to ignore the error and return None. Second is when openfunction can't open the file because
        # either the file isn't there or the permissions don't allow access.
        if errno.errorcode[err.errno] in ["ENOENT", "EACCES"]:
            six.reraise(*sys.exc_info())
        return None
    if mode.startswith("r"):
        # additional check for reading (eg can we uncompress) --- is this needed?
        try:
            stream.readline()
        except IOError:
            stream.close()
            stream = None
        except:
            stream.close()
            raise
        else:
            stream.close()
            stream = openfunction(filename, mode=mode)
    return stream


def hasmethod(obj, m):
    """Return ``True`` if object *obj* contains the method *m*.

    .. versionadded:: 0.7.1
    """
    return hasattr(obj, m) and callable(getattr(obj, m))


def isstream(obj):
    """Detect if `obj` is a stream.

    We consider anything a stream that has the methods

    - ``close()``

    and either set of the following

    - ``read()``, ``readline()``, ``readlines()``
    - ``write()``, ``writeline()``, ``writelines()``

    :Arguments:
      *obj*
          stream or str

    :Returns:
      *bool*, ``True`` if `obj` is a stream, ``False`` otherwise

    .. SeeAlso::
       :mod:`io`


    .. versionadded:: 0.7.1
    """
    signature_methods = ("close",)
    alternative_methods = (
        ("read", "readline", "readlines"),
        ("write", "writeline", "writelines"),
    )

    # Must have ALL the signature methods
    for m in signature_methods:
        if not hasmethod(obj, m):
            return False
    # Must have at least one complete set of alternative_methods
    alternative_results = [
        numpy.all([hasmethod(obj, m) for m in alternatives])
        for alternatives in alternative_methods
    ]
    return numpy.any(alternative_results)


# TODO: make it work for non-default charge state amino acids.
#: translation table for 1-letter codes --> 3-letter codes
#: .. Note: This does not work for HISB and non-default charge state aa!
amino_acid_codes = {
    "A": "ALA",
    "C": "CYS",
    "D": "ASP",
    "E": "GLU",
    "F": "PHE",
    "G": "GLY",
    "H": "HIS",
    "I": "ILE",
    "K": "LYS",
    "L": "LEU",
    "M": "MET",
    "N": "ASN",
    "P": "PRO",
    "Q": "GLN",
    "R": "ARG",
    "S": "SER",
    "T": "THR",
    "V": "VAL",
    "W": "TRP",
    "Y": "TYR",
}
inverse_aa_codes = {three: one for one, three in amino_acid_codes.items()}


def convert_aa_code(x):
    """Converts between 3-letter and 1-letter amino acid codes."""
    if len(x) == 1:
        return amino_acid_codes[x.upper()]
    elif len(x) == 3:
        return inverse_aa_codes[x.upper()]
    else:
        raise ValueError(
            "Can only convert 1-letter or 3-letter amino acid codes, " "not %r" % x
        )


@contextmanager
def in_dir(directory, create=True):
    """Context manager to execute a code block in a directory.

    * The directory is created if it does not exist (unless
      create=False is set)
    * At the end or after an exception code always returns to
      the directory that was the current directory before entering
      the block.
    """
    startdir = os.getcwd()
    try:
        try:
            os.chdir(directory)
            logger.debug("Working in {directory!r}...".format(**vars()))
        except OSError as err:
            if create and err.errno == errno.ENOENT:
                os.makedirs(directory)
                os.chdir(directory)
                logger.info(
                    "Working in {directory!r} (newly created)...".format(**vars())
                )
            else:
                logger.exception(
                    "Failed to start working in {directory!r}.".format(**vars())
                )
                raise
        yield os.getcwd()
    finally:
        os.chdir(startdir)


def realpath(*args):
    """Join all args and return the real path, rooted at /.

    Expands ``~`` and environment variables such as :envvar:`$HOME`.

    Returns ``None`` if any of the args is none.
    """
    if None in args:
        return None
    return os.path.realpath(os.path.expandvars(os.path.expanduser(os.path.join(*args))))


def find_first(filename, suffices=None):
    """Find first *filename* with a suffix from *suffices*.

    :Arguments:
      *filename*
         base filename; this file name is checked first
      *suffices*
         list of suffices that are tried in turn on the root of *filename*; can contain the
         ext separator (:data:`os.path.extsep`) or not

    :Returns: The first match or ``None``.
    """
    # struct is not reliable as it depends on qscript so now we just try everything...

    root, extension = os.path.splitext(filename)
    if suffices is None:
        suffices = []
    else:
        suffices = withextsep(suffices)
    extensions = [extension] + suffices  # native name is first
    for ext in extensions:
        fn = root + ext
        if os.path.exists(fn):
            return fn
    return None


def withextsep(extensions):
    """Return list in which each element is guaranteed to start with :data:`os.path.extsep`."""

    def dottify(x):
        if x.startswith(os.path.extsep):
            return x
        return os.path.extsep + x

    return [dottify(x) for x in asiterable(extensions)]


def find_files(directory, pattern):
    """Find files recursively under *directory*, matching *pattern* (generator).

    *pattern* is a UNIX-style glob pattern as used ny :func:`fnmatch.fnmatch`.

    Recipe by Bruno Oliveira from
    http://stackoverflow.com/questions/2186525/use-a-glob-to-find-files-recursively-in-python
    """
    for root, dirs, files in os.walk(directory):
        for basename in files:
            if fnmatch.fnmatch(basename, pattern):
                filename = os.path.join(root, basename)
                yield filename


def which(program):
    """Determine full path of executable *program* on :envvar:`PATH`.

    (Jay at http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python)

    .. versionadded:: 0.5.1
    """

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        real_program = realpath(program)
        if is_exe(real_program):
            return real_program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
    return None


class FileUtils(object):
    """Mixin class to provide additional file-related capabilities."""

    #: Default extension for files read/written by this class.
    default_extension = None

    def _init_filename(self, filename=None, ext=None):
        """Initialize the current filename :attr:`FileUtils.real_filename` of the object.

        Bit of a hack.

        - The first invocation must have ``filename != None``; this will set a
          default filename with suffix :attr:`FileUtils.default_extension`
          unless another one was supplied.

        - Subsequent invocations either change the filename accordingly or
          ensure that the default filename is set with the proper suffix.

        """

        extension = ext or self.default_extension
        filename = self.filename(
            filename, ext=extension, use_my_ext=True, set_default=True
        )
        #: Current full path of the object for reading and writing I/O.
        self.real_filename = os.path.realpath(filename)

    def filename(self, filename=None, ext=None, set_default=False, use_my_ext=False):
        """Supply a file name for the class object.

        Typical uses::

           fn = filename()             ---> <default_filename>
           fn = filename('name.ext')   ---> 'name'
           fn = filename(ext='pickle') ---> <default_filename>'.pickle'
           fn = filename('name.inp','pdf') --> 'name.pdf'
           fn = filename('foo.pdf',ext='png',use_my_ext=True) --> 'foo.pdf'

        The returned filename is stripped of the extension
        (``use_my_ext=False``) and if provided, another extension is
        appended. Chooses a default if no filename is given.

        Raises a ``ValueError`` exception if no default file name is known.

        If ``set_default=True`` then the default filename is also set.

        ``use_my_ext=True`` lets the suffix of a provided filename take
        priority over a default ``ext`` tension.

        .. versionchanged:: 0.3.1
           An empty string as *ext* = "" will suppress appending an extension.
        """
        if filename is None:
            if not hasattr(self, "_filename"):
                self._filename = None  # add attribute to class
            if self._filename:
                filename = self._filename
            else:
                raise ValueError(
                    "A file name is required because no default file name was defined."
                )
            my_ext = None
        else:
            filename, my_ext = os.path.splitext(filename)
            if set_default:  # replaces existing default file name
                self._filename = filename
        if my_ext and use_my_ext:
            ext = my_ext
        if ext is not None:
            if ext.startswith(os.extsep):
                ext = ext[1:]  # strip a dot to avoid annoying mistakes
            if ext != "":
                filename = filename + os.extsep + ext
        return filename

    def check_file_exists(self, filename, resolve="exception", force=None):
        """If a file exists then continue with the action specified in ``resolve``.

        ``resolve`` must be one of

        "ignore"
              always return ``False``
        "indicate"
              return ``True`` if it exists
        "warn"
              indicate and issue a :exc:`UserWarning`
        "exception"
              raise :exc:`IOError` if it exists

        Alternatively, set *force* for the following behaviour (which
        ignores *resolve*):

        ``True``
              same as *resolve* = "ignore" (will allow overwriting of files)
        ``False``
              same as *resolve* = "exception" (will prevent overwriting of files)
        ``None``
              ignored, do whatever *resolve* says
        """

        def _warn(x):
            msg = "File {0!r} already exists.".format(x)
            logger.warning(msg)
            warnings.warn(msg)
            return True

        def _raise(x):
            msg = "File {0!r} already exists.".format(x)
            logger.error(msg)
            raise IOError(errno.EEXIST, x, msg)

        solutions = {
            "ignore": lambda x: False,  # file exists, but we pretend that it doesn't
            "indicate": lambda x: True,  # yes, file exists
            "warn": _warn,
            "warning": _warn,
            "exception": _raise,
            "raise": _raise,
        }

        if force is True:
            resolve = "ignore"
        elif force is False:
            resolve = "exception"

        if not os.path.isfile(filename):
            return False
        else:
            return solutions[resolve](filename)

    def infix_filename(self, name, default, infix, ext=None):
        """Unless *name* is provided, insert *infix* before the extension *ext* of *default*."""
        if name is None:
            p, oldext = os.path.splitext(default)
            if ext is None:
                ext = oldext
            if ext.startswith(os.extsep):
                ext = ext[1:]
            name = self.filename(p + infix, ext=ext)
        return name

    def __repr__(self):
        fmt = "{0!s}(filename=%r)".format(self.__class__.__name__)
        try:
            fn = self.filename()
        except ValueError:
            fn = None
        return fmt % fn


def iterable(obj):
    """Returns ``True`` if *obj* can be iterated over and is *not* a  string."""
    if isinstance(obj, string_types):
        return False  # avoid iterating over characters of a string
    if hasattr(obj, "next"):
        return True  # any iterator will do
    try:
        len(obj)  # anything else that might work
    except TypeError:
        return False
    return True


def asiterable(obj):
    """Returns obj so that it can be iterated over; a string is *not* treated as iterable"""
    if not iterable(obj):
        obj = [obj]
    return obj


def firstof(obj):
    """Returns the first entry of a sequence or the obj.

    Treats strings as single objects.
    """
    return asiterable(obj)[0]


# In utilities so that it can be safely used in tools, cbook, ...


def unlink_f(path):
    """Unlink path but do not complain if file does not exist."""
    try:
        os.unlink(path)
    except OSError as err:
        if err.errno != errno.ENOENT:
            raise


def unlink_gmx(*args):
    """Unlink (remove) Gromacs file(s) and all corresponding backups."""
    for path in args:
        unlink_f(path)
    unlink_gmx_backups(*args)


def unlink_gmx_backups(*args):
    """Unlink (rm) all backup files corresponding to the listed files."""
    for path in args:
        dirname, filename = os.path.split(path)
        fbaks = glob.glob(os.path.join(dirname, "#" + filename + ".*#"))
        for bak in fbaks:
            unlink_f(bak)


def mkdir_p(path):
    """Create a directory *path* with subdirs but do not complain if it exists.

    This is like GNU ``mkdir -p path``.
    """
    try:
        os.makedirs(path)
    except OSError as err:
        if err.errno != errno.EEXIST:
            raise


def cat(f=None, o=None):
    """Concatenate files *f*=[...] and write to *o*"""
    # need f, o to be compatible with trjcat and eneconv
    if f is None or o is None:
        return
    target = o
    infiles = asiterable(f)
    logger.debug("cat {0!s} > {1!s} ".format(" ".join(infiles), target))
    with open(target, "w") as out:
        rc = subprocess.call(["cat"] + infiles, stdout=out)
    if rc != 0:
        msg = "failed with return code {0:d}: cat {1!r} > {2!r} ".format(
            rc, " ".join(infiles), target
        )
        logger.exception(msg)
        raise OSError(errno.EIO, msg, target)


# helpers for matplotlib
def activate_subplot(numPlot):
    """Make subplot *numPlot* active on the canvas.

    Use this if a simple ``subplot(numRows, numCols, numPlot)``
    overwrites the subplot instead of activating it.
    """
    # see http://www.mail-archive.com/matplotlib-users@lists.sourceforge.net/msg07156.html
    from pylab import gcf, axes

    numPlot -= 1  # index is 0-based, plots are 1-based
    return axes(gcf().get_axes()[numPlot])


def remove_legend(ax=None):
    """Remove legend for axes or gca.

    See http://osdir.com/ml/python.matplotlib.general/2005-07/msg00285.html
    """
    from pylab import gca, draw

    if ax is None:
        ax = gca()
    ax.legend_ = None
    draw()


# time functions
class Timedelta(datetime.timedelta):
    """Extension of :class:`datetime.timedelta`.

    Provides attributes ddays, dhours, dminutes, dseconds to measure
    the delta in normal time units.

    ashours gives the total time in fractional hours.
    """

    @property
    def dhours(self):
        """Hours component of the timedelta."""
        return self.seconds // 3600

    @property
    def dminutes(self):
        """Minutes component of the timedelta."""
        return self.seconds // 60 - 60 * self.dhours

    @property
    def dseconds(self):
        """Seconds component of the timedelta."""
        return self.seconds - 3600 * self.dhours - 60 * self.dminutes

    @property
    def ashours(self):
        """Timedelta in (fractional) hours."""
        return 24 * self.days + self.seconds / 3600.0

    def strftime(self, fmt="%d:%H:%M:%S"):
        """Primitive string formatter.

        The only directives understood are the following:
          ============   ==========================
          Directive      meaning
          ============   ==========================
          %d             day as integer
          %H             hour  [00-23]
          %h             hours including days
          %M             minute as integer [00-59]
          %S             second as integer [00-59]
          ============   ==========================
        """
        substitutions = {
            "%d": str(self.days),
            "%H": "{0:02d}".format(self.dhours),
            "%h": str(24 * self.days + self.dhours),
            "%M": "{0:02d}".format(self.dminutes),
            "%S": "{0:02d}".format(self.dseconds),
        }
        s = fmt
        for search, replacement in substitutions.items():
            s = s.replace(search, replacement)
        return s


NUMBERED_PDB = re.compile(r"(?P<PREFIX>.*\D)(?P<NUMBER>\d+)\.(?P<SUFFIX>pdb)")


def number_pdbs(*args, **kwargs):
    """Rename pdbs x1.pdb ... x345.pdb --> x0001.pdb ... x0345.pdb

    :Arguments:
       - *args*: filenames or glob patterns (such as "pdb/md*.pdb")
       - *format*: format string including keyword *num* ["%(num)04d"]
    """

    format = kwargs.pop("format", "%(num)04d")
    name_format = "%(prefix)s" + format + ".%(suffix)s"

    for f in itertools.chain.from_iterable(map(glob.glob, args)):
        m = NUMBERED_PDB.search(f)
        if m is None:
            continue
        num = int(m.group("NUMBER"))
        prefix = m.group("PREFIX")
        suffix = m.group("SUFFIX")
        newname = name_format % vars()
        logger.info("Renaming {f!r} --> {newname!r}".format(**vars()))
        try:
            os.rename(f, newname)
        except OSError:
            logger.exception("renaming failed")



# GromacsWrapper: top.py
# Copyright (c) 2009 Oliver Beckstein <orbeckst@gmail.com>
# Released under the GNU Public License 3 (or higher, your choice)
# See the file COPYING for details.

from __future__ import division, absolute_import, print_function

import os.path

import numpy as np

from numpy.testing import assert_array_equal

try:
    from pandas.testing import assert_frame_equal
except ImportError:
    # old versions of pandas
    from pandas.util.testing import assert_frame_equal

import pytest

import gromacs
from gromacs.fileformats import TOP, XVG
from gromacs import scaling

from ...datafiles import datafile


def errmsg_helper(attr, attr1, attr2):
    msg = ["attr = " + str(attr)]
    for pair in zip(attr1, attr2):
        if pair[0] == pair[1]:
            continue
        msg.append(str(pair))
    return "\n".join(msg)


def grompp(f, c, p, prefix="topol", **kwargs):
    s = prefix + ".tpr"
    po = prefix + ".mdp"

    rc, output, junk = gromacs.grompp(
        f=f, p=p, c=c, o=s, po=po, stdout=False, stderr=False, **kwargs
    )
    assert rc == 0, "grompp -o {0} -po {1} -f {2} -c {3} -p {4} failed".format(
        s, po, f, c, p
    )
    return s


def mdrun(s, prefix, nt=0):
    o = prefix + ".trr"
    rc, output, junk = gromacs.mdrun(
        v=True, s=s, o=o, stdout=False, stderr=False, nt=nt
    )
    assert rc == 0, "mdrun failed"
    return o


def rerun_energy(s, o, prefix, nt=0):
    e = prefix + ".edr"
    rc, output, junk = gromacs.mdrun(
        v=True, s=s, rerun=o, e=e, stdout=False, stderr=False, nt=nt
    )
    assert rc == 0, "mdrun failed"

    xvg = prefix + ".xvg"
    rc, output, junk = gromacs.g_energy(
        f=e,
        o=xvg,
        input=(
            "Proper-Dih.",
            "Improper-Dih.",
            "CMAP-Dih.",
            "LJ-14",
            "Coulomb-14",
            "LJ-(SR)",
            "Coulomb-(SR)",
            "Coul.-recip.",
            "Potential",
        ),
        stdout=False,
        stderr=False,
    )
    assert rc == 0, "g_energy failed"

    return XVG(xvg).to_df()


class TopologyTest(object):
    mdp = datafile("fileformats/top/grompp.mdp")

    def test_basic(self):
        path = self.processed
        top = TOP(path)
        assert list(top.dict_molname_mol.keys()) == self.molecules

    def test_equal(self):
        """
        Load the same topology twice and check if __eq__ comparisions work
        """
        path = self.processed

        top1 = TOP(path)
        top2 = TOP(path)

        attrs1 = [section for section in top1.found_sections if "types" in section]
        attrs2 = [section for section in top1.found_sections if "types" in section]

        assert attrs1 == attrs2, "sections are not equal"

        for attr in attrs1:
            assert getattr(top1, attr) == getattr(
                top2, attr
            ), "{0} not identical".format(attr)

    # def test_parameter_types(self):
    #         """Test if all the parameter types are the same across two topologies
    #         """
    #         pass

    # def test_parameters(self):
    #         """Test if per-molecule parameters and parameter type assignments are identical
    #         """
    #         pass

    # def test_molecule_parameters(self):
    #         """Called by `test_parameters()` for each molecule in the system
    #         """

    def test_read(self, tmpdir):
        path = self.processed
        try:
            top1 = TOP(path)
        except Exception as err:
            raise AssertionError(
                "Failed to read {0}. Raised:\n{1}".format(path, str(err))
            )
        attrs1 = [section for section in top1.found_sections if "types" in section]
        assert attrs1

    def test_read_write(self, tmpdir):
        """Read a topology, write it out, and read in the output again.
        Writing the topology out should make no change to the topology.
        """
        path = self.processed
        with tmpdir.as_cwd():
            filename1 = "processed_1.top"
            filename2 = "processed_2.top"

            top1 = TOP(path)
            top1.write(filename1)

            # make life harder, write out again
            top2 = TOP(filename1)
            top2.write(filename2)

            top2 = TOP(filename2)

            attrs1 = [section for section in top1.found_sections if "types" in section]
            attrs2 = [section for section in top1.found_sections if "types" in section]

            assert attrs1 == attrs2

            # attrs1 = ['atomtypes', 'pairtypes', 'bondtypes', 'constrainttypes', 'angletypes', 'dihedraltypes', 'dihedraltypes', 'cmaptypes']
            # attrs1 = ['atomtypes',]

            for attr in attrs1:
                attr1 = getattr(top1, attr)
                attr2 = getattr(top2, attr)
                assert attr1 == attr2, errmsg_helper(attr, attr1, attr2)

    def test_grompp(self, tmpdir):
        """Check if grompp can be run successfully at all"""
        f = self.mdp
        c = self.conf
        p = self.processed
        with tmpdir.as_cwd():
            o = "topol.tpr"
            po = "mdout.mdp"
            rc, output, junk = gromacs.grompp(
                f=f, p=p, c=c, o=o, po=po, stdout=False, stderr=False
            )
            assert rc == 0, "grompp -f {0} -o {1} ... failed to run".format(f, o)

    def test_mdrun(self, tmpdir, low_performance):
        """Check if grompp can be run successfully at all"""
        # set low_performance with
        #
        #    pytest --low-performance gromacs/tests
        #
        f = self.mdp
        c = self.conf
        processed = self.processed

        nt = 2 if low_performance else 0

        with tmpdir.as_cwd():
            tpr = grompp(f, c, processed, prefix="reference")
            reference_trr = mdrun(tpr, prefix="reference", nt=nt)
            df1 = rerun_energy(tpr, reference_trr, prefix="reference", nt=nt)

            scaled = "scaled.top"
            kwargs = dict(
                banned_lines="",
                topfile=processed,
                outfile=scaled,
                scale_lipids=1.0,
                scale_protein=1.0,
            )
            scaling.partial_tempering(**kwargs)

            assert os.path.exists(scaled), "failed to produce {0}".format(scaled)

            tpr = grompp(f, c, scaled, prefix="scaled")
            df2 = rerun_energy(tpr, reference_trr, prefix="scaled", nt=nt)

            assert_frame_equal(df1, df2, check_names=True, check_like=True)

            scaled = "scaled.top"
            kwargs = dict(
                banned_lines="",
                topfile=processed,
                outfile=scaled,
                scale_lipids=1.0,
                scale_protein=0.5,
            )
            scaling.partial_tempering(**kwargs)
            tpr = grompp(f, c, scaled, prefix="scaled", maxwarn=1)
            df3 = rerun_energy(tpr, reference_trr, prefix="scaled", nt=nt)
            # print(df1, df1.columns)
            # print(df3, df3.columns)
            unscaled_terms = ["Time (ps)", "Improper Dih."]
            scaled_terms = ["Proper Dih."]

            assert_frame_equal(
                df1[unscaled_terms],
                df3[unscaled_terms],
                check_names=True,
                check_like=True,
            )
            assert_frame_equal(
                df1[scaled_terms],
                2 * df3[scaled_terms],
                check_names=True,
                check_like=True,
            )



# setuptools installation of GromacsWrapper
# Copyright (c) 2008-2011 Oliver Beckstein <orbeckst@gmail.com>
# Released under the GNU Public License 3 (or higher, your choice)
#
# See the files INSTALL and README for details or visit
# https://github.com/Becksteinlab/GromacsWrapper
from __future__ import with_statement
from setuptools import setup, find_packages

import versioneer

with open("README.rst") as readme:
    long_description = readme.read()


setup(
    name="GromacsWrapper",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="A Python wrapper around the Gromacs tools.",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author="Oliver Beckstein",
    author_email="orbeckst@gmail.com",
    license="GPLv3",
    url="https://github.com/Becksteinlab/GromacsWrapper",
    download_url="https://github.com/Becksteinlab/GromacsWrapper/downloads",
    keywords="science Gromacs analysis 'molecular dynamics'",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows ",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    packages=find_packages(
        exclude=["scripts", "tests", "tests.*", "extras", "doc/examples"]
    ),
    scripts=[
        "scripts/gw-join_parts.py",
        "scripts/gw-merge_topologies.py",
        "scripts/gw-forcefield.py",
        "scripts/gw-partial_tempering.py",
    ],
    package_data={
        "gromacs": [
            "templates/*.sge",
            "templates/*.pbs",  # template files
            "templates/*.ll",
            "templates/*.sh",
            "templates/*.mdp",
            "templates/*.cfg",
        ],
    },
    install_requires=[
        "numpy>=1.0",
        "six",  # towards py 3 compatibility
        "numkit",  # numerical helpers
        "matplotlib",
    ],
    tests_require=["pytest", "pandas>=0.17"],
    zip_safe=True,
)



# GromacsWrapper: formats.py
# Copyright (c) 2009-2011 Oliver Beckstein <orbeckst@gmail.com>
# Released under the GNU Public License 3 (or higher, your choice)
# See the file COPYING for details.
"""
Gromacs parameter MDP file format
=================================

The `.mdp file`_ contains a list of keywords that are used to set up a
simulation with :class:`~gromacs.tools.Grompp`. The class :class:`MDP`
parses this file and provides access to the keys and values as ordered
dictionary.

.. _`.mdp file`: http://www.gromacs.org/Documentation/File_Formats/.mdp_File

.. autoclass:: MDP
   :members:
"""


from __future__ import absolute_import, with_statement

import os, errno
import re
import warnings
import six
import numpy


from ..exceptions import ParseError, AutoCorrectionWarning
from .. import utilities
from collections import OrderedDict as odict

import logging


class MDP(odict, utilities.FileUtils):
    """Class that represents a Gromacs mdp run input file.

    The MDP instance is an ordered dictionary.

      - *Parameter names* are keys in the dictionary.
      - *Comments* are sequentially numbered with keys Comment0001,
        Comment0002, ...
      - *Empty lines* are similarly preserved as Blank0001, ....

    When writing, the dictionary is dumped in the recorded order to a
    file. Inserting keys at a specific position is not possible.

    Currently, comments after a parameter on the same line are
    discarded. Leading and trailing spaces are always stripped.

    .. SeeAlso:: For editing a mdp file one can also use
                :func:`gromacs.cbook.edit_mdp` (which works like a
                poor replacement for sed).
    """

    default_extension = "mdp"
    logger = logging.getLogger("gromacs.formats.MDP")

    COMMENT = re.compile("""\s*;\s*(?P<value>.*)""")  # eat initial ws
    # see regex in cbook.edit_mdp()
    PARAMETER = re.compile(
        """
                            \s*(?P<parameter>[^=]+?)\s*=\s*  # parameter (ws-stripped), before '='
                            (?P<value>[^;]*)                # value (stop before comment=;)
                            (?P<comment>\s*;.*)?            # optional comment
                            """,
        re.VERBOSE,
    )

    def __init__(self, filename=None, autoconvert=True, **kwargs):
        """Initialize mdp structure.

        :Arguments:
          *filename*
              read from mdp file
          *autoconvert* : boolean
              ``True`` converts numerical values to python numerical types;
              ``False`` keeps everything as strings [``True``]
          *kwargs*
              Populate the MDP with key=value pairs. (NO SANITY CHECKS; and also
              does not work for keys that are not legal python variable names such
              as anything that includes a minus '-' sign or starts with a number).
        """
        super(MDP, self).__init__(
            **kwargs
        )  # can use kwargs to set dict! (but no sanity checks!)

        self.autoconvert = autoconvert

        if filename is not None:
            self._init_filename(filename)
            self.read(filename)

    def _transform(self, value):
        if self.autoconvert:
            return utilities.autoconvert(value)
        else:
            return value.rstrip()

    def read(self, filename=None):
        """Read and parse mdp file *filename*."""
        self._init_filename(filename)

        def BLANK(i):
            return "B{0:04d}".format(i)

        def COMMENT(i):
            return "C{0:04d}".format(i)

        data = odict()
        iblank = icomment = 0
        with open(self.real_filename) as mdp:
            for line in mdp:
                line = line.strip()
                if len(line) == 0:
                    iblank += 1
                    data[BLANK(iblank)] = ""
                    continue
                m = self.COMMENT.match(line)
                if m:
                    icomment += 1
                    data[COMMENT(icomment)] = m.group("value")
                    continue
                # parameter
                m = self.PARAMETER.match(line)
                if m:
                    # check for comments after parameter?? -- currently discarded
                    parameter = m.group("parameter")
                    value = self._transform(m.group("value"))
                    data[parameter] = value
                else:
                    errmsg = "{filename!r}: unknown line in mdp file, {line!r}".format(
                        **vars()
                    )
                    self.logger.error(errmsg)
                    raise ParseError(errmsg)

        super(MDP, self).update(data)

    def write(self, filename=None, skipempty=False):
        """Write mdp file to *filename*.

        :Keywords:
           *filename*
               output mdp file; default is the filename the mdp
               was read from
           *skipempty* : boolean
               ``True`` removes any parameter lines from output that
               contain empty values [``False``]

        .. Note:: Overwrites the file that the mdp was read from if no
                  *filename* supplied.
        """

        with open(self.filename(filename, ext="mdp"), "w") as mdp:
            for k, v in self.items():
                if k[0] == "B":  # blank line
                    mdp.write("\n")
                elif k[0] == "C":  # comment
                    mdp.write("; {v!s}\n".format(**vars()))
                else:  # parameter = value
                    if skipempty and (v == "" or v is None):
                        continue
                    if isinstance(v, six.string_types) or not hasattr(v, "__iter__"):
                        mdp.write("{k!s} = {v!s}\n".format(**vars()))
                    else:
                        mdp.write("{} = {}\n".format(k, " ".join(map(str, v))))



# GromacsWrapper: test_example.py
# Copyright (c) 2009 Oliver Beckstein <orbeckst@gmail.com>
# Released under the GNU Public License 3 (or higher, your choice)
# See the file COPYING for details.

import pytest

import gromacs


def test_version():
    release = gromacs.__version__
    assert isinstance(release, str)



# GromacsWrapper: scaling.py
# Released under the GNU Public License 3 (or higher, your choice)
# See the file COPYING for details.

"""

:mod:`gromacs.scaling` -- Partial tempering
===========================================

:Author: Jan Domanski, @jandom

.. versionadded:: 0.5.0

Helper functions for scaling gromacs topologies; useful for setting up
simulations with Hamiltonian replicate exchange and partial tempering
(REST2).

.. autofunction:: scale_dihedrals
.. autofunction:: scale_impropers
.. autofunction:: partial_tempering

"""
from __future__ import absolute_import, division, print_function

import math
import copy
import logging

import numpy as np

from .fileformats import TOP
from .fileformats import blocks

logger = logging.getLogger("gromacs.scaling")


def scale_dihedrals(mol, dihedrals, scale, banned_lines=None):
    """Scale dihedral angles"""

    if banned_lines is None:
        banned_lines = []
    new_dihedrals = []
    for dh in mol.dihedrals:
        atypes = (
            dh.atom1.get_atomtype(),
            dh.atom2.get_atomtype(),
            dh.atom3.get_atomtype(),
            dh.atom4.get_atomtype(),
        )
        atypes = [a.replace("_", "").replace("=", "") for a in atypes]

        # special-case: this is a [ dihedral ] override in molecule block, continue and don't match
        if dh.gromacs["param"] != []:
            for p in dh.gromacs["param"]:
                p["kch"] *= scale
            new_dihedrals.append(dh)
            continue

        for iswitch in range(32):
            if iswitch % 2 == 0:
                a1 = atypes[0]
                a2 = atypes[1]
                a3 = atypes[2]
                a4 = atypes[3]
            else:
                a1 = atypes[3]
                a2 = atypes[2]
                a3 = atypes[1]
                a4 = atypes[0]

            if (iswitch // 2) % 2 == 1:
                a1 = "X"
            if (iswitch // 4) % 2 == 1:
                a2 = "X"
            if (iswitch // 8) % 2 == 1:
                a3 = "X"
            if (iswitch // 16) % 2 == 1:
                a4 = "X"
            key = "{0}-{1}-{2}-{3}-{4}".format(a1, a2, a3, a4, dh.gromacs["func"])
            if key in dihedrals:
                for i, dt in enumerate(dihedrals[key]):
                    dhA = copy.deepcopy(dh)
                    param = copy.deepcopy(dt.gromacs["param"])
                    # Only check the first dihedral in a list
                    if not dihedrals[key][0].line in banned_lines:
                        for p in param:
                            p["kchi"] *= scale
                    dhA.gromacs["param"] = param
                    # if key == "CT3-C-NH1-CT1-9": print i, dt, key
                    if i == 0:
                        dhA.comment = "; banned lines {0} found={1}\n".format(
                            " ".join(map(str, banned_lines)),
                            1 if dt.line in banned_lines else 0,
                        )
                        dhA.comment += (
                            "; parameters for types {}-{}-{}-{}-9 at LINE({})\n".format(
                                dhA.atom1.atomtype,
                                dhA.atom2.atomtype,
                                dhA.atom3.atomtype,
                                dhA.atom4.atomtype,
                                dt.line,
                            ).replace("_", "")
                        )
                    name = "{}-{}-{}-{}-9".format(
                        dhA.atom1.atomtype,
                        dhA.atom2.atomtype,
                        dhA.atom3.atomtype,
                        dhA.atom4.atomtype,
                    ).replace("_", "")
                    # if name == "CL-CTL2-CTL2-HAL2-9": print dihedrals[key], key
                    new_dihedrals.append(dhA)
                break

    mol.dihedrals = new_dihedrals
    # assert(len(mol.dihedrals) == new_dihedrals)
    return mol


def scale_impropers(mol, impropers, scale, banned_lines=None):
    """Scale improper dihedrals"""
    if banned_lines is None:
        banned_lines = []
    new_impropers = []
    for im in mol.impropers:
        atypes = (
            im.atom1.get_atomtype(),
            im.atom2.get_atomtype(),
            im.atom3.get_atomtype(),
            im.atom4.get_atomtype(),
        )
        atypes = [a.replace("_", "").replace("=", "") for a in atypes]

        # special-case: this is a [ dihedral ] override in molecule block, continue and don't match
        if im.gromacs["param"] != []:
            for p in im.gromacs["param"]:
                p["kpsi"] *= scale
            new_impropers.append(im)
            continue

        for iswitch in range(32):
            if iswitch % 2 == 0:
                a1 = atypes[0]
                a2 = atypes[1]
                a3 = atypes[2]
                a4 = atypes[3]
            else:
                a1 = atypes[3]
                a2 = atypes[2]
                a3 = atypes[1]
                a4 = atypes[0]
            if (iswitch // 2) % 2 == 1:
                a1 = "X"
            if (iswitch // 4) % 2 == 1:
                a2 = "X"
            if (iswitch // 8) % 2 == 1:
                a3 = "X"
            if (iswitch // 16) % 2 == 1:
                a4 = "X"
            key = "{0}-{1}-{2}-{3}-{4}".format(a1, a2, a3, a4, im.gromacs["func"])
            if key in impropers:
                for i, imt in enumerate(impropers[key]):
                    imA = copy.deepcopy(im)
                    param = copy.deepcopy(imt.gromacs["param"])
                    # Only check the first dihedral in a list
                    if not impropers[key][0].line in banned_lines:
                        for p in param:
                            p["kpsi"] *= scale
                    imA.gromacs["param"] = param
                    if i == 0:
                        imA.comment = "; banned lines {0} found={1}\n ; parameters for types {2}-{3}-{4}-{5}-9 at LINE({6})\n".format(
                            " ".join(map(str, banned_lines)),
                            1 if imt.line in banned_lines else 0,
                            imt.atype1,
                            imt.atype2,
                            imt.atype3,
                            imt.atype4,
                            imt.line,
                        )
                    new_impropers.append(imA)
                break
    # assert(len(mol.impropers) == new_impropers)
    mol.impropers = new_impropers
    return mol


def partial_tempering(
    topfile="processed.top",
    outfile="scaled.top",
    banned_lines="",
    scale_lipids=1.0,
    scale_protein=1.0,
):
    """Set up topology for partial tempering (REST2) replica exchange.


    .. versionchanged:: 0.7.0
       Use keyword arguments instead of an `args` Namespace object.
    """

    banned_lines = map(int, banned_lines.split())
    top = TOP(topfile)
    groups = [("_", float(scale_protein)), ("=", float(scale_lipids))]

    #
    # CMAPTYPES
    #
    cmaptypes = []
    for ct in top.cmaptypes:
        cmaptypes.append(ct)
        for gr, scale in groups:
            ctA = copy.deepcopy(ct)
            ctA.atype1 += gr
            ctA.atype2 += gr
            ctA.atype3 += gr
            ctA.atype4 += gr
            ctA.atype8 += gr
            ctA.gromacs["param"] = [v * scale for v in ct.gromacs["param"]]
            cmaptypes.append(ctA)
    logger.debug("cmaptypes was {0}, is {1}".format(len(top.cmaptypes), len(cmaptypes)))
    top.cmaptypes = cmaptypes

    #
    # ATOMTYPES
    #
    atomtypes = []
    for at in top.atomtypes:
        atomtypes.append(at)
        for gr, scale in groups:
            atA = copy.deepcopy(at)
            atA.atnum = atA.atype
            atA.atype += gr
            atA.gromacs["param"]["lje"] *= scale
            atomtypes.append(atA)
    top.atomtypes = atomtypes

    #
    # PAIRTYPES
    #
    pairtypes = []
    for pt in top.pairtypes:
        pairtypes.append(pt)
        for gr, scale in groups:
            ptA = copy.deepcopy(pt)
            ptA.atype1 += gr
            ptA.atype2 += gr
            ptA.gromacs["param"]["lje14"] *= scale

            pairtypes.append(ptA)
    top.pairtypes = pairtypes

    #
    # BONDTYPES
    #
    bondtypes = []
    for bt in top.bondtypes:
        bondtypes.append(bt)
        for gr, scale in groups:
            btA = copy.deepcopy(bt)
            btA.atype1 += gr
            btA.atype2 += gr
            bondtypes.append(btA)
    top.bondtypes = bondtypes

    #
    # ANGLETYPES
    #
    angletypes = []
    for at in top.angletypes:
        angletypes.append(at)
        for gr, scale in groups:
            atA = copy.deepcopy(at)
            atA.atype1 += gr
            atA.atype2 += gr
            atA.atype3 += gr
            angletypes.append(atA)
    top.angletypes = angletypes

    #
    # Build dihedral dictionary
    #
    dihedraltypes = {}
    for dt in top.dihedraltypes:
        dt.disabled = True
        dt.comment = "; type={0!s}-{1!s}-{2!s}-{3!s}-9\n; LINE({4:d}) ".format(
            dt.atype1, dt.atype2, dt.atype3, dt.atype4, dt.line
        )
        dt.comment = dt.comment.replace("_", "")

        # if "X-CTL2-CTL2-X-9" in dt.comment: print dt
        name = "{0}-{1}-{2}-{3}-{4}".format(
            dt.atype1, dt.atype2, dt.atype3, dt.atype4, dt.gromacs["func"]
        )
        if not name in dihedraltypes:
            dihedraltypes[name] = []
        dihedraltypes[name].append(dt)
    logger.debug(
        "Build dihedraltypes dictionary with {0} entries".format(len(dihedraltypes))
    )

    #
    # Build improper dictionary
    #
    impropertypes = {}
    for it in top.impropertypes:
        it.disabled = True
        it.comment = "; LINE({0:d}) ".format(it.line)
        name = "{0}-{1}-{2}-{3}-{4}".format(
            it.atype1, it.atype2, it.atype3, it.atype4, it.gromacs["func"]
        )
        if not name in impropertypes:
            impropertypes[name] = []
        impropertypes[name].append(it)
    logger.debug(
        "Build impropertypes dictionary with {0} entries".format(len(impropertypes))
    )

    for molname_mol in top.dict_molname_mol:
        if not "Protein" in molname_mol:
            continue
        mol = top.dict_molname_mol[molname_mol]
        for at in mol.atoms:
            at.charge *= math.sqrt(scale_protein)
        mol = scale_dihedrals(mol, dihedraltypes, scale_protein, banned_lines)
        mol = scale_impropers(mol, impropertypes, 1.0, banned_lines)

    top.write(outfile)



# GromacsWrapper: test_example.py
# Copyright (c) 2009 Oliver Beckstein <orbeckst@gmail.com>
# Released under the GNU Public License 3 (or higher, your choice)
# See the file COPYING for details.

from __future__ import division, absolute_import, print_function

import os.path
import pytest
from six.moves import cPickle as pickle
from six.moves import StringIO
from six.moves.collections_abc import Iterable

import numpy as np

import gromacs.utilities


@pytest.fixture
def string_buffer():
    return StringIO()


def test_which(name="cat"):
    path = gromacs.utilities.which(name)
    assert os.path.basename(path) == name


@pytest.mark.parametrize("path", ["~/whatever", "$HOME/whatever"])
def test_realpath(path):
    abspath = gromacs.utilities.realpath(path)
    assert abspath.startswith(os.path.sep)
    assert abspath.startswith(os.path.expanduser("~"))


class TestAttributeDict(object):
    def setup(self):
        self.d = gromacs.utilities.AttributeDict(foo="bar", baz="boing")

    def test_attribute_get(self):
        assert self.d.foo == "bar"

    def test_dict_get(self):
        assert self.d["foo"] == "bar"
        assert self.d.get("foo") == "bar"

    def test_attribute_set(self):
        self.d.gargl = "blaster"
        assert self.d.gargl == "blaster"

    def test_dict_set(self):
        self.d["gargl"] = "blaster"
        assert self.d["gargl"] == "blaster"

    def test_pickle(self):
        try:
            dump = pickle.dumps(self.d, pickle.HIGHEST_PROTOCOL)
        except Exception as err:
            raise AssertionError("serializing failed: {}".format(err))
        try:
            d = pickle.loads(dump)
        except Exception as err:
            raise AssertionError("de-serializing failed: {0}".format(str(err)))
        assert set(d.keys()) == set(self.d.keys())
        assert set(d.values()) == set(self.d.values())


@pytest.fixture(
    params=[
        (42, int),
        ("42", int),
        ([42], int),
        (2.7, float),
        ("2.7", float),
        ([2.7], float),
        ("jabberwock", str),
        (["foo", "bar"], str),
        ("42 42", np.integer),
        ("2.7 2.7", float),
        ("foo bar", str),
    ],
    ids=[
        "int -> int",
        "str -> int",
        "int list -> int list",
        "float -> float",
        "str -> float",
        "float list -> float list",
        "str -> str",
        "str list -> str list",
        "str -> int list",
        "str -> float list",
        "str -> str list",
    ],
)
def conversions(request):
    value, target_type = request.param
    x = gromacs.utilities.autoconvert(value)
    return x, value, target_type


def test_autoconvert(conversions):
    x, value, target_type = conversions
    if isinstance(x, Iterable):
        x = x[0]
    assert isinstance(x, target_type), "Failed to convert '{0}' to type {1}".format(
        value, target_type
    )


@pytest.fixture(params=["", "gz", "bz2"])
def openanyfilename(request):
    return "lifeofbrian.txt." + request.param


class TestOpenAny(object):
    quote = """There shall, in that time, be rumours of things going astray,
    erm, and there shall be a great confusion as to where things really are,
    and nobody will really know where lieth those little things... with the
    sort of raffia work base that has an attachment. At this time, a friend
    shall lose his friend's hammer and the young shall not know where lieth
    the things possessed by their fathers that their fathers put there only
    just the night before, about eight o'clock."""

    def test_file(self, tmpdir, openanyfilename):
        filename = openanyfilename
        name, ext = os.path.splitext(filename)
        if ext in [".gz", ".bz2"]:
            quote = self.quote.encode("utf8")
        else:
            quote = self.quote

        with tmpdir.as_cwd():
            with gromacs.utilities.openany(filename, "w") as datafile:
                datafile.write(quote)
            with gromacs.utilities.openany(filename, "r") as datafile:
                data = datafile.read()
        assert data == quote

    def test_stream_write(self, string_buffer):
        with gromacs.utilities.openany(string_buffer, "w") as datafile:
            datafile.write(self.quote)
            # note: closes stream so not easily useable with StringIO (see NamedStream
            # in MDAnalysis)
            # check inside with block
            string_buffer.seek(0)
            assert string_buffer.read() == self.quote

    def test_stream_read(self, string_buffer):
        string_buffer.write(self.quote)
        string_buffer.seek(0)
        with gromacs.utilities.openany(string_buffer, "r") as datafile:
            data = datafile.read()
            # note: closes stream so not easily useable with StringIO (see NamedStream
            # in MDAnalysis)
            # check inside with block
            assert data == self.quote


@pytest.fixture
def pdb_files(tmpdir):
    for i in [1, 15, 300]:
        tmpdir.join("myfile{}.pdb".format(i)).write("foo\nbar\n")

    currdir = os.getcwd()
    try:
        os.chdir(str(tmpdir))
        yield
    finally:
        os.chdir(currdir)


@pytest.mark.parametrize(
    "args",
    [
        ("myfile*.pdb",),
        ("myfile1.pdb", "myfile15.pdb", "myfile300.pdb"),
        ("myfile1.pdb", "myfile*.pdb"),
        ("myfile*.pdb", "myotherfiles*.pdb"),
    ],
)
def test_number_pdbs(pdb_files, args):
    gromacs.utilities.number_pdbs(*args)

    assert os.path.exists("myfile0001.pdb")
    assert os.path.exists("myfile0015.pdb")
    assert os.path.exists("myfile0300.pdb")


def test_cat(tmpdir):
    # assert cat.noise == miaow
    # assert not cat.noise == woof
    f1 = tmpdir.join("file1.txt")
    f1.write("foo\n")
    f2 = tmpdir.join("file2.txt")
    f2.write("bar\n")
    out = tmpdir.join("out.txt")
    gromacs.utilities.cat([str(f1), str(f2)], str(out))

    assert out.read() == "foo\nbar\n"


def test_cat_fail(tmpdir):
    with pytest.raises(OSError):
        gromacs.utilities.cat(["not_here.txt"], str(tmpdir.join("new.txt")))


def test_cat_None_out(tmpdir):
    gromacs.utilities.cat(["some_file.txt"], None)


def test_cat_None_in(tmpdir):
    gromacs.utilities.cat(None, str(tmpdir.join("out.txt")))

    assert not os.path.exists(str(tmpdir.join("out.txt")))


@pytest.fixture
def unlink_files(tmpdir):
    tmpdir.join("hello.mdp").write("foo\n")
    tmpdir.join("#hello.mdp.1#").write("foo\n")
    tmpdir.join("#hello.mdp.2#").write("foo\n")

    tmpdir.join("out.gro").write("bar\n")
    tmpdir.join("#out.gro.1#").write("bar\n")
    tmpdir.join("#out.gro.2#").write("bar\n")

    currdir = os.getcwd()
    try:
        os.chdir(str(tmpdir))
        yield
    finally:
        os.chdir(currdir)


def test_unlink(unlink_files):
    assert os.path.exists("out.gro")

    gromacs.utilities.unlink_f("out.gro")

    assert not os.path.exists("out.gro")


def test_unlink_nonexistant(unlink_files):
    assert not os.path.exists("out.xtc")
    gromacs.utilities.unlink_f("out.xtc")


def test_unlink_gmx_backups(unlink_files):
    gromacs.utilities.unlink_gmx_backups("hello.mdp")

    assert os.path.exists("hello.mdp")
    assert not os.path.exists("#hello.mdp.1#")
    assert not os.path.exists("#hello.mdp.2#")
    assert os.path.exists("out.gro")
    assert os.path.exists("#out.gro.1#")


def test_unlink_gmx(unlink_files):
    gromacs.utilities.unlink_gmx("hello.mdp")
    assert not os.path.exists("hello.mdp")
    assert not os.path.exists("#hello.mdp.1#")
    assert not os.path.exists("#hello.mdp.2#")
    assert os.path.exists("out.gro")
    assert os.path.exists("#out.gro.1#")


@pytest.mark.parametrize(
    "iterable,expected",
    [
        ("this", "this"),
        (["this", "that"], "this"),
        ([1, 2, 3], 1),
        (np.arange(4), 0),
    ],
)
def test_firstof(iterable, expected):
    assert gromacs.utilities.firstof(iterable) == expected


@pytest.mark.parametrize(
    "val,ref",
    [
        ("a", "ALA"),
        ("A", "ALA"),
        ("ala", "A"),
        ("ALA", "A"),
        ("Ala", "A"),
        ("Q", "GLN"),
        ("q", "GLN"),
        ("GLN", "Q"),
        ("gln", "Q"),
        ("Gln", "Q"),
    ],
)
def test_conv_aa_code(val, ref):
    assert gromacs.utilities.convert_aa_code(val) == ref


@pytest.mark.parametrize("val", ["ALAA", ""])
def test_conv_aa_code_VE(val):
    with pytest.raises(ValueError):
        gromacs.utilities.convert_aa_code(val)


@pytest.fixture
def fileutil():
    class MyFileUtil(gromacs.utilities.FileUtils):
        default_extension = ".test"

        def __init__(self):
            self._init_filename(filename="simple.test")

    return MyFileUtil()


@pytest.mark.parametrize(
    "filename,ext,ref",
    [
        (None, None, "simple"),
        ("other", None, "other"),
        (None, "pdf", "simple.pdf"),
        ("other", "pdf", "other.pdf"),
    ],
)
def test_FileUtils_filename(fileutil, filename, ext, ref):
    assert fileutil.filename(filename=filename, ext=ext) == ref


def test_FileUtils_filename_VE(fileutil):
    del fileutil._filename

    with pytest.raises(ValueError):
        fileutil.filename()


@pytest.fixture
def fileutil_withfiles(fileutil, tmpdir):
    curr = os.getcwd()

    tmpdir.join("exists.txt").write("hello\n")

    try:
        os.chdir(str(tmpdir))
        yield fileutil
    finally:
        os.chdir(curr)


@pytest.mark.parametrize("fn", ["exists.txt", "nonexistant.txt"])
def test_check_file_exists_ignore(fileutil_withfiles, fn):
    assert fileutil_withfiles.check_file_exists(fn, resolve="ignore") is False


@pytest.mark.parametrize("fn", ["exists.txt", "nonexistant.txt"])
def test_check_file_exists_force(fileutil_withfiles, fn):
    assert fileutil_withfiles.check_file_exists(fn, force=True) is False


@pytest.mark.parametrize("fn,ref", [("exists.txt", True), ("nonexistant.txt", False)])
def test_check_file_exists_indicate(fileutil_withfiles, fn, ref):
    assert fileutil_withfiles.check_file_exists(fn, resolve="indicate") is ref


@pytest.mark.parametrize("resolve", ["warn", "warning"])
def test_check_file_exists_warn(fileutil_withfiles, resolve):
    with pytest.warns(UserWarning):
        fileutil_withfiles.check_file_exists("exists.txt", resolve=resolve)


@pytest.mark.parametrize("resolve", ["exception", "raise"])
def test_check_file_exists_raise(fileutil_withfiles, resolve):
    with pytest.raises(IOError):
        fileutil_withfiles.check_file_exists("exists.txt", resolve=resolve)
